@@ -1,11 +1,11 @@
 /*
- * $Id: clext.c 220 2020-01-01 10:17:21Z vruppert $
+ * $Id: clext.c 292 2021-06-03 12:24:22Z vruppert $
  */
 //
 //  QEMU Cirrus CLGD 54xx VGABIOS Extension.
 //
 //  Copyright (c) 2004      Makoto Suzuki (suzu)
-//  Copyright (C) 2004-2020 Volker Ruppert
+//  Copyright (C) 2004-2021 Volker Ruppert
 //
 //  This library is free software; you can redistribute it and/or
 //  modify it under the terms of the GNU Lesser General Public
@@ -26,10 +26,6 @@
 #undef CIRRUS_VESA3_PMINFO
 #endif
 
-#define PM_BIOSMEM_CURRENT_MODE 0x449
-#define PM_BIOSMEM_CRTC_ADDRESS 0x463
-#define PM_BIOSMEM_VBE_MODE 0x4BA
-
 typedef struct
 {
   /* + 0 */
@@ -45,17 +41,9 @@ typedef struct
   /* +16 */
   unsigned char bitsperpixel;
   unsigned char vesacolortype;
-  unsigned char vesaredmask;
-  unsigned char vesaredpos;
-  unsigned char vesagreenmask;
-  unsigned char vesagreenpos;
-  unsigned char vesabluemask;
-  unsigned char vesabluepos;
-  /* +24 */
-  unsigned char vesareservedmask;
-  unsigned char vesareservedpos;
+  unsigned char *color_params;
 } cirrus_mode_t;
-#define CIRRUS_MODE_SIZE 26
+#define CIRRUS_MODE_SIZE 20
 
 
 /* For VESA BIOS 3.0 */
@@ -75,9 +63,7 @@ unsigned short cgraph_svgacolor[] = {
 /* 640x480x8 */
 unsigned short cseq_640x480x8[] = {
 0x0300,0x2101,0x0f02,0x0003,0x0e04,0x1107,
-0x580b,0x580c,0x580d,0x580e,
 0x0412,0x0013,0x2017,
-0x331b,0x331c,0x331d,0x331e,
 0xffff
 };
 unsigned short ccrtc_640x480x8[] = {
@@ -91,9 +77,7 @@ unsigned short ccrtc_640x480x8[] = {
 /* 640x480x16 */
 unsigned short cseq_640x480x16[] = {
 0x0300,0x2101,0x0f02,0x0003,0x0e04,0x1707,
-0x580b,0x580c,0x580d,0x580e,
 0x0412,0x0013,0x2017,
-0x331b,0x331c,0x331d,0x331e,
 0xffff
 };
 unsigned short ccrtc_640x480x16[] = {
@@ -107,9 +91,7 @@ unsigned short ccrtc_640x480x16[] = {
 /* 640x480x24 */
 unsigned short cseq_640x480x24[] = {
 0x0300,0x2101,0x0f02,0x0003,0x0e04,0x1507,
-0x580b,0x580c,0x580d,0x580e,
 0x0412,0x0013,0x2017,
-0x331b,0x331c,0x331d,0x331e,
 0xffff
 };
 unsigned short ccrtc_640x480x24[] = {
@@ -122,14 +104,13 @@ unsigned short ccrtc_640x480x24[] = {
 };
 /* 800x600x8 */
 unsigned short cseq_800x600x8[] = {
-0x0300,0x2101,0x0f02,0x0003,0x0e04,0x1107,
-0x230b,0x230c,0x230d,0x230e,
-0x0412,0x0013,0x2017,
-0x141b,0x141c,0x141d,0x141e,
+0x0300,0x2101,0x0f02,0x0003,0x0e04,
+0x0412,0x0013,0x2017,0x510e,0x3a1e,
+0x1107,
 0xffff
 };
 unsigned short ccrtc_800x600x8[] = {
-0x2311,0x7d00,0x6301,0x6302,0x8003,0x6b04,0x1a05,0x9806,0xf007,
+0x2311,0x7f00,0x6301,0x6402,0x8203,0x6b04,0x1b05,0x7206,0xf007,
 0x6009,0x000c,0x000d,
 0x7d10,0x5712,0x6413,0x4014,0x5715,0x9816,0xc317,0xff18,
 0x001a,0x221b,0x001d,
@@ -137,14 +118,13 @@ unsigned short ccrtc_800x600x8[] = {
 };
 /* 800x600x16 */
 unsigned short cseq_800x600x16[] = {
-0x0300,0x2101,0x0f02,0x0003,0x0e04,0x1707,
-0x230b,0x230c,0x230d,0x230e,
-0x0412,0x0013,0x2017,
-0x141b,0x141c,0x141d,0x141e,
+0x0300,0x2101,0x0f02,0x0003,0x0e04,
+0x0412,0x0013,0x2017,0x510e,0x3a1e,
+0x1707,
 0xffff
 };
 unsigned short ccrtc_800x600x16[] = {
-0x2311,0x7d00,0x6301,0x6302,0x8003,0x6b04,0x1a05,0x9806,0xf007,
+0x2311,0x7f00,0x6301,0x6402,0x8203,0x6b04,0x1b05,0x7206,0xf007,
 0x6009,0x000c,0x000d,
 0x7d10,0x5712,0xc813,0x4014,0x5715,0x9816,0xc317,0xff18,
 0x001a,0x221b,0x001d,
@@ -152,14 +132,13 @@ unsigned short ccrtc_800x600x16[] = {
 };
 /* 800x600x24 */
 unsigned short cseq_800x600x24[] = {
-0x0300,0x2101,0x0f02,0x0003,0x0e04,0x1507,
-0x230b,0x230c,0x230d,0x230e,
-0x0412,0x0013,0x2017,
-0x141b,0x141c,0x141d,0x141e,
+0x0300,0x2101,0x0f02,0x0003,0x0e04,
+0x0412,0x0013,0x2017,0x510e,0x3a1e,
+0x1507,
 0xffff
 };
 unsigned short ccrtc_800x600x24[] = {
-0x2311,0x7d00,0x6301,0x6302,0x8003,0x6b04,0x1a05,0x9806,0xf007,
+0x2311,0x7f00,0x6301,0x6402,0x8203,0x6b04,0x1b05,0x7206,0xf007,
 0x6009,0x000c,0x000d,
 0x7d10,0x5712,0x2c13,0x4014,0x5715,0x9816,0xc317,0xff18,
 0x001a,0x321b,0x001d,
@@ -167,10 +146,9 @@ unsigned short ccrtc_800x600x24[] = {
 };
 /* 1024x768x8 */
 unsigned short cseq_1024x768x8[] = {
-0x0300,0x2101,0x0f02,0x0003,0x0e04,0x1107,
-0x760b,0x760c,0x760d,0x760e,
-0x0412,0x0013,0x2017,
-0x341b,0x341c,0x341d,0x341e,
+0x0300,0x2101,0x0f02,0x0003,0x0e04,
+0x0412,0x0013,0x2017,0x760e,0x341e,
+0x1107,
 0xffff
 };
 unsigned short ccrtc_1024x768x8[] = {
@@ -182,10 +160,9 @@ unsigned short ccrtc_1024x768x8[] = {
 };
 /* 1024x768x16 */
 unsigned short cseq_1024x768x16[] = {
-0x0300,0x2101,0x0f02,0x0003,0x0e04,0x1707,
-0x760b,0x760c,0x760d,0x760e,
-0x0412,0x0013,0x2017,
-0x341b,0x341c,0x341d,0x341e,
+0x0300,0x2101,0x0f02,0x0003,0x0e04,
+0x0412,0x0013,0x2017,0x760e,0x341e,
+0x1707,
 0xffff
 };
 unsigned short ccrtc_1024x768x16[] = {
@@ -197,10 +174,9 @@ unsigned short ccrtc_1024x768x16[] = {
 };
 /* 1024x768x24 */
 unsigned short cseq_1024x768x24[] = {
-0x0300,0x2101,0x0f02,0x0003,0x0e04,0x1507,
-0x760b,0x760c,0x760d,0x760e,
-0x0412,0x0013,0x2017,
-0x341b,0x341c,0x341d,0x341e,
+0x0300,0x2101,0x0f02,0x0003,0x0e04,
+0x0412,0x0013,0x2017,0x760e,0x341e,
+0x1507,
 0xffff
 };
 unsigned short ccrtc_1024x768x24[] = {
@@ -212,111 +188,117 @@ unsigned short ccrtc_1024x768x24[] = {
 };
 /* 1280x1024x8 */
 unsigned short cseq_1280x1024x8[] = {
-0x0300,0x2101,0x0f02,0x0003,0x0e04,0x1107,
-0x760b,0x760c,0x760d,0x760e,
-0x0412,0x0013,0x2017,
-0x341b,0x341c,0x341d,0x341e,
+0x0300,0x2101,0x0f02,0x0003,0x0e04,
+0x0412,0x0013,0x2017,0x6e0e,0x2a1e,
+0x1107,
 0xffff
 };
 unsigned short ccrtc_1280x1024x8[] = {
-0x2911,0xc300,0x9f01,0x9f02,0x8603,0x8304,0x9405,0x2406,0xf707,
+0x2911,0xbd00,0x9f01,0xa002,0x8003,0xa404,0x1905,0x2a06,0xb207,
 0x6009,0x000c,0x000d,
-0x0310,0xff12,0xa013,0x4014,0xff15,0x2416,0xc317,0xff18,
-0x001a,0x221b,0x001d,
+0x0b10,0xff12,0xa013,0x4014,0xff15,0x2416,0xc317,0xff18,
+0x011a,0x221b,0x001d,
 0xffff
 };
 /* 1280x1024x16 */
 unsigned short cseq_1280x1024x16[] = {
-0x0300,0x2101,0x0f02,0x0003,0x0e04,0x1707,
-0x760b,0x760c,0x760d,0x760e,
-0x0412,0x0013,0x2017,
-0x341b,0x341c,0x341d,0x341e,
+0x0300,0x2101,0x0f02,0x0003,0x0e04,
+0x0412,0x0013,0x2017,0x06e0e,0x2a1e,
+0x1707,
 0xffff
 };
 unsigned short ccrtc_1280x1024x16[] = {
-0x2911,0xc300,0x9f01,0x9f02,0x8603,0x8304,0x9405,0x2406,0xf707,
+0x2911,0xbd00,0x9f01,0xa002,0x8003,0xa404,0x1905,0x2a06,0xb207,
 0x6009,0x000c,0x000d,
-0x0310,0xff12,0x4013,0x4014,0xff15,0x2416,0xc317,0xff18,
-0x001a,0x321b,0x001d,
+0x0b10,0xff12,0x4013,0x4014,0xff15,0x2416,0xc317,0xff18,
+0x011a,0x321b,0x001d,
 0xffff
 };
 
 /* 1600x1200x8 */
 unsigned short cseq_1600x1200x8[] = {
-0x0300,0x2101,0x0f02,0x0003,0x0e04,0x1107,
-0x760b,0x760c,0x760d,0x760e,
-0x0412,0x0013,0x2017,
-0x341b,0x341c,0x341d,0x341e,
+0x0300,0x2101,0x0f02,0x0003,0x0e04,
+0x0412,0x0013,0x2017,0x880e,0x221e,
+0x1107,
 0xffff
 };
 unsigned short ccrtc_1600x1200x8[] = {
-0x2911,0xc300,0x9f01,0x9f02,0x8603,0x8304,0x9405,0x2406,0xf707,
+0x2911,0xe800,0xc701,0xc802,0x8603,0x8304,0x9405,0x7206,0xf007,
 0x6009,0x000c,0x000d,
-0x0310,0xff12,0xa013,0x4014,0xff15,0x2416,0xc317,0xff18,
-0x001a,0x221b,0x001d,
+0x7d10,0x5712,0xc813,0x4014,0xff15,0x2416,0xc317,0xff18,
+0x011a,0x221b,0x001d,
 0xffff
 };
 
+unsigned char cirrus_color_params_8bpp[] = {
+0, 0, 0, 0, 0, 0, 0, 0};
+unsigned char cirrus_color_params_15bpp[] = {
+5, 10, 5, 5, 5, 0, 1, 15};
+unsigned char cirrus_color_params_16bpp[] = {
+5, 11, 6, 5, 5, 0, 0, 0};
+unsigned char cirrus_color_params_24bpp[] = {
+8, 16, 8, 8, 8, 0, 0, 0};
+
 cirrus_mode_t cirrus_modes[] =
 {
  {0x5f,640,480,8,0x00,
    cseq_640x480x8,cgraph_svgacolor,ccrtc_640x480x8,8,
-   4,0,0,0,0,0,0,0,0},
+   4,cirrus_color_params_8bpp},
  {0x64,640,480,16,0xe1,
    cseq_640x480x16,cgraph_svgacolor,ccrtc_640x480x16,16,
-   6,5,11,6,5,5,0,0,0},
+   6,cirrus_color_params_16bpp},
  {0x66,640,480,15,0xf0,
    cseq_640x480x16,cgraph_svgacolor,ccrtc_640x480x16,16,
-   6,5,10,5,5,5,0,1,15},
+   6,cirrus_color_params_15bpp},
  {0x71,640,480,24,0xe5,
    cseq_640x480x24,cgraph_svgacolor,ccrtc_640x480x24,24,
-   6,8,16,8,8,8,0,0,0},
+   6,cirrus_color_params_24bpp},
 
  {0x5c,800,600,8,0x00,
    cseq_800x600x8,cgraph_svgacolor,ccrtc_800x600x8,8,
-   4,0,0,0,0,0,0,0,0},
+   4,cirrus_color_params_8bpp},
  {0x65,800,600,16,0xe1,
    cseq_800x600x16,cgraph_svgacolor,ccrtc_800x600x16,16,
-   6,5,11,6,5,5,0,0,0},
+   6,cirrus_color_params_16bpp},
  {0x67,800,600,15,0xf0,
    cseq_800x600x16,cgraph_svgacolor,ccrtc_800x600x16,16,
-   6,5,10,5,5,5,0,1,15},
+   6,cirrus_color_params_15bpp},
 
  {0x60,1024,768,8,0x00,
    cseq_1024x768x8,cgraph_svgacolor,ccrtc_1024x768x8,8,
-   4,0,0,0,0,0,0,0,0},
+   4,cirrus_color_params_8bpp},
  {0x74,1024,768,16,0xe1,
    cseq_1024x768x16,cgraph_svgacolor,ccrtc_1024x768x16,16,
-   6,5,11,6,5,5,0,0,0},
+   6,cirrus_color_params_16bpp},
  {0x68,1024,768,15,0xf0,
    cseq_1024x768x16,cgraph_svgacolor,ccrtc_1024x768x16,16,
-   6,5,10,5,5,5,0,1,15},
+   6,cirrus_color_params_15bpp},
 
  {0x78,800,600,24,0xe5,
    cseq_800x600x24,cgraph_svgacolor,ccrtc_800x600x24,24,
-   6,8,16,8,8,8,0,0,0},
+   6,cirrus_color_params_24bpp},
  {0x79,1024,768,24,0xe5,
    cseq_1024x768x24,cgraph_svgacolor,ccrtc_1024x768x24,24,
-   6,8,16,8,8,8,0,0,0},
+   6,cirrus_color_params_24bpp},
 
  {0x6d,1280,1024,8,0x00,
    cseq_1280x1024x8,cgraph_svgacolor,ccrtc_1280x1024x8,8,
-   4,0,0,0,0,0,0,0,0},
+   4,cirrus_color_params_8bpp},
  {0x69,1280,1024,15,0xf0,
    cseq_1280x1024x16,cgraph_svgacolor,ccrtc_1280x1024x16,16,
-   6,5,10,5,5,5,0,1,15},
+   6,cirrus_color_params_15bpp},
  {0x75,1280,1024,16,0xe1,
    cseq_1280x1024x16,cgraph_svgacolor,ccrtc_1280x1024x16,16,
-   6,5,11,6,5,5,0,0,0},
+   6,cirrus_color_params_16bpp},
 
  {0x7b,1600,1200,8,0x00,
    cseq_1600x1200x8,cgraph_svgacolor,ccrtc_1600x1200x8,8,
-   4,0,0,0,0,0,0,0,0},
+   4,cirrus_color_params_8bpp},
 
  {0xfe,0,0,0,0,cseq_vga,cgraph_vga,ccrtc_vga,0,
-   0xff,0,0,0,0,0,0,0,0},
+   0xff,cirrus_color_params_8bpp},
  {0xff,0,0,0,0,0,0,0,0,
-   0xff,0,0,0,0,0,0,0,0},
+   0xff,0},
 };
 
 unsigned char cirrus_id_table[] = {
@@ -407,15 +389,7 @@ cirrus_init:
   out dx, ax
   mov ax, #0x0031
   out dx, ax
-  mov dx, #0x3C4
-  mov ax, #0x4a0b ; VCLK #0 setup
-  out dx, ax
-  mov ax, #0x2b1b
-  out dx, ax
-  mov ax, #0x5b0c ; VCLK #1 setup
-  out dx, ax
-  mov ax, #0x2f1c
-  out dx, ax
+  call cirrus_set_vga_vclk
   mov ax, #0x420d ; VCLK #2 setup
   out dx, ax
   mov ax, #0x1f1d
@@ -583,8 +557,23 @@ cirrus_set_video_mode_extended:
   pop ax
 cirrus_set_video_mode_extended_1:
   and al, #0x7f
+  call cirrus_set_video_mode_bda
+  SET_INT_VECTOR(0x43, #0xC000, #_vgafont16)
+  mov al, #0x20
+  pop si
+  jmp cirrus_return
 
+cirrus_set_video_mode_bda:
+  push bx
+  push cx
   push ds
+  push cs
+  pop ds
+  mov bx, [si+2]
+  shr bx, #3
+  mov cx, [si+4]
+  shr cx, #4 ; cheight 16
+  dec cx
 #ifdef CIRRUS_VESA3_PMINFO
  db 0x2e ;; cs:
   mov si, [cirrus_vesa_sel0000_data]
@@ -593,12 +582,17 @@ cirrus_set_video_mode_extended_1:
 #endif
   mov ds, si
   mov [PM_BIOSMEM_CURRENT_MODE], al
+  mov [PM_BIOSMEM_NB_COLS], bx
+  mov [PM_BIOSMEM_NB_ROWS], cl
+  mov al, #0x10
+  xor ah, ah
+  mov [PM_BIOSMEM_CHAR_HEIGHT], ax
+  xor ax, ax
+  mov [PM_BIOSMEM_CURSOR_POS], ax
   pop ds
-
-  mov al, #0x20
-
-  pop si
-  jmp cirrus_return
+  pop cx
+  pop bx
+  ret
 
 cirrus_vesa_pmbios_init:
   retf
@@ -634,6 +628,8 @@ cirrus_switch_mode:
   push cs
   pop ds
 
+  call cirrus_set_vga_vclk
+
   mov bx, [si+10] ;; seq
   mov dx, #0x3c4
   mov ax, #0x1206
@@ -645,9 +641,23 @@ cirrus_switch_mode:
   call cirrus_switch_mode_setregs
 
   mov bx, [si+14] ;; crtc
-  call cirrus_get_crtc
+  call get_crtc_address
   call cirrus_switch_mode_setregs
 
+  mov  bx, [si+2]
+  or   bx, bx
+  jz   no_vclk_setup
+  mov  dx, #VGAREG_READ_MISC_OUTPUT
+  in   al, dx
+  and  al, #0xf3
+  cmp  bx, #0x280
+  jbe  set_vclk_sel
+  or   al, #0x0c
+set_vclk_sel:
+  mov  dx, #VGAREG_WRITE_MISC_OUTPUT
+  out  dx, al
+no_vclk_setup:
+
   mov dx, #0x3c6
   mov al, #0x00
   out dx, al
@@ -716,9 +726,243 @@ csms_1:
 csms_2:
   ret
 
+cirrus_set_vga_vclk:
+  mov dx, #0x3C4
+  mov ax, #0x4a0b ; VCLK #0 setup
+  out dx, ax
+  mov ax, #0x2b1b
+  out dx, ax
+  mov ax, #0x5b0c ; VCLK #1 setup
+  out dx, ax
+  mov ax, #0x2f1c
+  out dx, ax
+  ret
+
+; code for 'write character' support in 8bpp graphics modes
+
+; called from C code
+_cirrus_is_8bpp_mode:
+cirrus_is_8bpp_mode:
+  push dx
+  mov  dx, #0x03c4
+  mov  al, #0x07
+  out  dx, al
+  inc  dx
+  in   al, dx
+  pop  dx
+  and  al, #0x0f
+  cmp  al, #0x01
+  jz   is_cirrus_8bpp
+  xor  ax, ax
+is_cirrus_8bpp:
+  ret
+
+cirrus_pm_get_bda_word:
+  push ds
+  push bx
+  push si
+#ifdef CIRRUS_VESA3_PMINFO
+ db 0x2e ;; cs:
+  mov  si, [cirrus_vesa_sel0000_data]
+#else
+  xor  si, si
+#endif
+  mov  ds, si
+  mov  bh, #0x04
+  mov  bl, al
+  mov  ax, [bx]
+  pop  si
+  pop  bx
+  pop  ds
+  ret
+
+; in: regnum in BL, value in AX
+cirrus_write_bltreg16:
+  push dx
+  push ax
+  mov  ah, al
+  mov  al, bl
+  mov  dx, #0x03ce
+  out  dx, ax
+  pop  ax
+  push ax
+  inc  bl
+  mov  al, bl
+  out  dx, ax
+  pop  ax
+  pop  dx
+  ret
+
+; in: columns in AL rows in BL / out: size of line in AX
+cirrus_write_bltsize:
+  xor  ah, ah
+  shl  ax, #3
+  dec  ax
+  push bx
+  mov  bl, #0x20
+  call cirrus_write_bltreg16
+  pop  bx
+  mov  ax, 14[bp] ; cheight
+  push ax
+  mul  bl
+  dec  ax
+  mov  bl, #0x22
+  call cirrus_write_bltreg16
+  mov  ax, 12[bp] ; nbrows
+  shl  ax, #3
+  mov  bl, #0x24
+  call cirrus_write_bltreg16
+  mov  bl, #0x26
+  call cirrus_write_bltreg16
+  pop bx
+  mul bx
+  ret
+
+; in: size of line in AX, ypos in BL, regnum in BH, xpos in CL
+cirrus_write_bltaddr:
+  push ax
+  push bx
+  xor  bh, bh
+  mul  bx
+  mov  bl, cl
+  shl  bx, #3
+  add  ax, bx
+  jnc  cirrus_addr_noinc
+  inc  dx
+cirrus_addr_noinc:
+  pop  bx
+  mov  bl, bh
+  call cirrus_write_bltreg16
+  mov  ax, dx
+  mov  ah, al
+  mov  al, bh
+  inc  al
+  inc  al
+  mov  dx, #0x03ce
+  out  dx, ax
+  pop  ax
+  ret
+
+cirrus_bitblt_write_char:
+  push ax
+  push bx
+  push cx
+  push dx
+  push ds
+  push es
+  push si
+  push di
+  xor  ax, ax
+  mov  ds, ax
+  mov  bx, #0x10c ; INT 0x43
+  mov  si, [bx]
+  mov  ax, [bx+2]
+  mov  ds, ax
+  mov  bx, 14[bp] ; cheight
+  mov  al, 4[bp]  ; character
+  mul  bl
+  add  si, ax
+  mov  cl, 8[bp]  ; xcurs
+  mov  ch, 10[bp] ; ycurs
+  mov  al, #0x01
+  mov  bl, #0x01
+  call cirrus_write_bltsize
+  mov  bl, ch
+  mov  bh, #0x28
+  call cirrus_write_bltaddr
+  mov  ax, #0x0001 ; src pitch 1
+  mov  bl, #0x26
+  call cirrus_write_bltreg16
+  mov  dx, #0x03ce
+  mov  ax, #0x8430 ; colorexpand memsrc
+  out  dx, ax
+  mov  ax, #0x0d32 ; ROP: copy
+  out  dx, ax
+  xor  ax, ax ; background
+  out  dx, ax
+  mov  ah, 6[bp] ; attr
+  mov  al, #0x01
+  out  dx, ax
+  mov  ax, #0x0231 ; start
+  out  dx, ax
+  mov  ax, #0xa000
+  mov  es, ax
+  xor  di, di
+  mov  cx, 14[bp] ; cheight
+  cld
+  rep
+    movsb
+  pop  di
+  pop  si
+  pop  es
+  pop  ds
+  pop  dx
+  pop  cx
+  pop  bx
+  pop  ax
+  ret
+
+cirrus_bitblt_copy:
+  push ax
+  push bx
+  push cx
+  push dx
+  mov  cl, 4[bp] ; xstart
+  mov  ch, 8[bp] ; ydest
+  mov  al, 10[bp] ; cols
+  mov  bl, #0x01
+  call cirrus_write_bltsize
+  mov  bl, 6[bp] ; ysrc
+  mov  bh, #0x2c
+  call cirrus_write_bltaddr
+  mov  bl, ch
+  mov  bh, #0x28
+  call cirrus_write_bltaddr
+  mov  dx, #0x03ce
+  mov  ax, #0x0030 ; normal blt
+  out  dx, ax
+  mov  ax, #0x0d32 ; ROP: copy
+  out  dx, ax
+  mov  ax, #0x0231 ; start
+  out  dx, ax
+  pop  dx
+  pop  cx
+  pop  bx
+  pop  ax
+  ret
+
+cirrus_bitblt_fill:
+  push ax
+  push bx
+  push dx
+  mov  cl, 4[bp] ; xstart
+  mov  ch, 6[bp] ; ydest
+  mov  al, 8[bp] ; cols
+  mov  bl, 10[bp] ; rows
+  call cirrus_write_bltsize
+  mov  bl, ch
+  mov  bh, #0x28
+  call cirrus_write_bltaddr
+  mov  dx, #0x03ce
+  mov  ax, #0x0030 ; normal blt
+  out  dx, ax
+  mov  ax, #0x0e32 ; ROP: 1
+  out  dx, ax
+  mov  ax, #0x0433 ; solidfill
+  out  dx, ax
+  mov  ah, 16[bp] ; attr
+  mov  al, #0x01
+  out  dx, ax
+  mov  ax, #0x0231
+  out  dx, ax
+  pop  dx
+  pop  bx
+  pop  ax
+  ret
+
 cirrus_extbios_80h:
   push dx
-  call cirrus_get_crtc
+  call get_crtc_address
   mov al, #0x27
   out dx, al
   inc dx
@@ -747,7 +991,7 @@ cirrus_extbios_81h:
   ret
 cirrus_extbios_82h:
   push dx
-  call cirrus_get_crtc
+  call get_crtc_address
   xor ax, ax
   mov al, #0x27
   out dx, al
@@ -918,9 +1162,16 @@ cirrus_vesa_01h_1:
     stosw ;; clear buffer
   pop di
 
-  mov ax, #0x003b ;; mode
+  mov  ax, #0x003b ;; mode attr
+  push bx
+  mov  bl, [si+6] ;; bpp
+  cmp  bl, #0x08
+  jne  cirrus_vesa_no_8bpp
+  or   al, #0x04 ;; TTY support
+cirrus_vesa_no_8bpp:
+  pop  bx
   stosw
-  mov ax, #0x0007 ;; attr
+  mov ax, #0x0007 ;; win attr
   stosw
   mov ax, #0x0010 ;; granularity =16K
   stosw
@@ -972,13 +1223,14 @@ cirrus_vesa_01h_3:
   stosb
 
   ;; v1.2+ stuffs
+  mov  ax, [si+18]
   push si
-  add si, #18
+  mov  si, ax
   movsw
   movsw
   movsw
   movsw
-  pop si
+  pop  si
 
   mov ah, [si+16]
   mov al, #0x0
@@ -1018,20 +1270,19 @@ cirrus_vesa_01h_4:
   pop ds
 
   test cx, #0x4000 ;; LFB flag
-  jz cirrus_vesa_01h_5
+  jz   cirrus_vesa_01h_5
   push cx
  db 0x26 ;; es:
-  mov cx, [di]
-  cmp cx, #0x0080 ;; is LFB supported?
-  jnz cirrus_vesa_01h_6
-  mov ax, #0x014F ;; error - no LFB
+  mov  cx, [di]
+  test cx, #0x0080 ;; is LFB supported?
+  jnz  cirrus_vesa_01h_6
+  mov  ax, #0x014F ;; error - no LFB
 cirrus_vesa_01h_6:
-  pop cx
+  pop  cx
 cirrus_vesa_01h_5:
   ret
 
 cirrus_vesa_02h:
-  ;; XXX support CRTC registers
   test bx, #0x3e00
   jnz cirrus_vesa_02h_2 ;; unknown flags
   mov ax, bx
@@ -1042,12 +1293,18 @@ cirrus_vesa_02h:
   cmp ax, #0xffff
   jnz cirrus_vesa_02h_1
 cirrus_vesa_02h_2:
-  jmp cirrus_vesa_unimplemented
+  jmp  cirrus_vesa_unimplemented
 cirrus_vesa_02h_legacy:
+  test al, #0x80
+  jnz  cirrus_vesa_02h_2
+  test bx, #0x8000
+  jz   cirrus_vesa_02h_legacy_noclear
+  or   al, #0x80
+cirrus_vesa_02h_legacy_noclear:
 #ifdef CIRRUS_VESA3_PMINFO
  db 0x2e ;; cs:
-  cmp byte ptr [cirrus_vesa_is_protected_mode], #0
-  jnz cirrus_vesa_02h_2
+  cmp  byte ptr [cirrus_vesa_is_protected_mode], #0
+  jnz  cirrus_vesa_02h_2
 #endif // CIRRUS_VESA3_PMINFO
   int #0x10
   mov ax, #0x004F
@@ -1063,12 +1320,12 @@ cirrus_vesa_02h_1:
 cirrus_vesa_02h_3:
   test bx, #0x8000 ;; no clear
   jnz cirrus_vesa_02h_4
-  push ax
-  xor ax,ax
+  xor ax, ax
   call cirrus_clear_vram
-  pop ax
 cirrus_vesa_02h_4:
   pop ax
+  call cirrus_set_video_mode_bda
+  SET_INT_VECTOR(0x43, #0xC000, #_vgafont16)
   push ds
 #ifdef CIRRUS_VESA3_PMINFO
  db 0x2e ;; cs:
@@ -1077,7 +1334,6 @@ cirrus_vesa_02h_4:
   xor si, si
 #endif
   mov ds, si
-  mov [PM_BIOSMEM_CURRENT_MODE], al
   mov [PM_BIOSMEM_VBE_MODE], bx
   pop ds
   pop si
@@ -1444,7 +1700,7 @@ cirrus_vesa_read_EDID:
   call cirrus_ddc_delay
   mov  al, #0xa1
   call cirrus_ddc_send_byte
-  jnz  cirrus_vesa_unimplemented
+  jc   cirrus_vesa_unimplemented
   push cx
   push di
   mov  cx, #0x0080
@@ -1488,21 +1744,6 @@ cvtm_2:
   pop ds
   ret
 
-  ; cirrus_get_crtc
-  ;; NOTE - may be called in protected mode
-cirrus_get_crtc:
-  push ds
-  push ax
-  mov  dx, #0x3cc
-  in   al, dx
-  and  al, #0x01
-  shl  al, #5
-  mov  dx, #0x3b4
-  add  dl, al
-  pop  ax
-  pop  ds
-  ret
-
 ;; in - al:mode, out - cflag:result, si:table, ax:destroyed
 cirrus_get_modeentry:
   and al, #0x7f
@@ -1551,7 +1792,7 @@ cirrus_get_bpp_bytes_2:
 cirrus_set_line_offset:
   shr  ax, #3
   push ax
-  call cirrus_get_crtc
+  call get_crtc_address
   mov  al, #0x13
   out  dx, al
   inc  dx
@@ -1563,7 +1804,7 @@ cirrus_set_line_offset:
   inc  dx
   shl  ah, #4
   in   al, dx
-  and  al, #ef
+  and  al, #0xef
   or   al, ah
   out  dx, al
   ret
@@ -1572,7 +1813,7 @@ cirrus_set_line_offset:
 cirrus_get_line_offset:
   push dx
   push bx
-  call cirrus_get_crtc
+  call get_crtc_address
   mov  al, #0x13
   out  dx, al
   inc  dx
@@ -1630,7 +1871,7 @@ cirrus_set_start_addr:
   push bx
   push dx
   push ax
-  call cirrus_get_crtc
+  call get_crtc_address
   mov  al, #0x0d
   out  dx, al
   inc  dx
@@ -1673,7 +1914,7 @@ cirrus_set_start_addr:
 ;; out - current address in DX:AX
 cirrus_get_start_addr:
   push bx
-  call cirrus_get_crtc
+  call get_crtc_address
   mov  al, #0x0c
   out  dx, al
   inc  dx
