@@ -1,10 +1,10 @@
 /*
- * $Id: vbe.c 228 2020-01-02 23:09:02Z vruppert $
+ * $Id: vbe.c 292 2021-06-03 12:24:22Z vruppert $
  */
 // ============================================================================================
 //
 //  Copyright (C) 2002      Jeroen Janssen
-//  Copyright (C) 2003-2020 Volker Ruppert
+//  Copyright (C) 2003-2021 Volker Ruppert
 //
 //  This library is free software; you can redistribute it and/or
 //  modify it under the terms of the GNU Lesser General Public
@@ -22,8 +22,8 @@
 //
 // ============================================================================================
 //
-//  This VBE is part of the VGA Bios specific to the plex86/bochs Emulated VGA card. 
-//  You can NOT drive any physical vga card with it. 
+//  This VBE is part of the VGA Bios specific to the plex86/bochs Emulated VGA card.
+//  You can NOT drive any physical vga card with it.
 //
 // ============================================================================================
 //
@@ -53,7 +53,7 @@ extern char vbebios_product_revision;
 ASM_START
 // FIXME: 'merge' these (c) etc strings with the vgabios.c strings?
 _vbebios_copyright:
-.ascii       "Bochs VBE (C) 2002-2020 http://savannah.nongnu.org/projects/vgabios/"
+.ascii       "Bochs VBE (C) 2002-2021 http://savannah.nongnu.org/projects/vgabios/"
 .byte        0x00
 
 _vbebios_vendor_name:
@@ -65,7 +65,7 @@ _vbebios_product_name:
 .byte        0x00
 
 _vbebios_product_revision:
-.ascii       "$Id: vbe.c 228 2020-01-02 23:09:02Z vruppert $"
+.ascii       "$Id: vbe.c 292 2021-06-03 12:24:22Z vruppert $"
 .byte        0x00
 
 _vbebios_info_string:
@@ -82,7 +82,7 @@ _no_vbebios_info_string:
 
 #if defined(USE_BX_INFO) || defined(DEBUG)
 msg_vbe_init:
-.ascii      "VBE Bios $Id: vbe.c 228 2020-01-02 23:09:02Z vruppert $"
+.ascii      "VBE Bios $Id: vbe.c 292 2021-06-03 12:24:22Z vruppert $"
 .byte	0x0a,0x0d, 0x00
 #endif
 
@@ -94,9 +94,7 @@ vesa_pm_start:
   dw vesa_pm_io_ports_table - vesa_pm_start
 vesa_pm_io_ports_table:
   dw VBE_DISPI_IOPORT_INDEX
-  dw VBE_DISPI_IOPORT_INDEX + 1
   dw VBE_DISPI_IOPORT_DATA
-  dw VBE_DISPI_IOPORT_DATA + 1
   dw 0xffff
   dw 0xffff
 
@@ -134,7 +132,7 @@ vesa_pm_set_display_start:
   mov  ax, #0x0100
   ret
 vesa_pm_set_display_start1:
-; convert offset to (X, Y) coordinate 
+; convert offset to (X, Y) coordinate
 ; (would be simpler to change Bochs VBE API...)
   push eax
   push ecx
@@ -243,12 +241,8 @@ dispi_set_id:
   out  dx, ax
   pop  dx
   ret
-ASM_END
 
-static void dispi_set_xres(xres)
-  Bit16u xres;
-{
-ASM_START
+_dispi_set_xres:
   push bp
   mov  bp, sp
   push ax
@@ -264,24 +258,44 @@ ASM_START
   pop  dx
   pop  ax
   pop  bp
-ASM_END
-}
+  ret
 
-static void dispi_set_yres(yres)
-  Bit16u yres;
-{
-  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_YRES);
-  outw(VBE_DISPI_IOPORT_DATA,yres);
-}
+_dispi_set_yres:
+  push bp
+  mov  bp, sp
+  push ax
+  push dx
 
-static void dispi_set_bpp(bpp)
-  Bit16u bpp;
-{
-  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_BPP);
-  outw(VBE_DISPI_IOPORT_DATA,bpp);
-}
+  mov  dx, # VBE_DISPI_IOPORT_INDEX
+  mov  ax, # VBE_DISPI_INDEX_YRES
+  out  dx, ax
+  mov  dx, # VBE_DISPI_IOPORT_DATA
+  mov  ax, 4[bp] ; yres
+  out  dx, ax
+
+  pop  dx
+  pop  ax
+  pop  bp
+  ret
+
+_dispi_set_bpp:
+  push bp
+  mov  bp, sp
+  push ax
+  push dx
+
+  mov  dx, # VBE_DISPI_IOPORT_INDEX
+  mov  ax, # VBE_DISPI_INDEX_BPP
+  out  dx, ax
+  mov  dx, # VBE_DISPI_IOPORT_DATA
+  mov  ax, 4[bp] ; bpp
+  out  dx, ax
+
+  pop  dx
+  pop  ax
+  pop  bp
+  ret
 
-ASM_START
 ; AL = bits per pixel / AH = bytes per pixel
 dispi_get_bpp:
   push dx
@@ -361,6 +375,28 @@ _dispi_get_max_bpp:
   pop  dx
   ret
 
+_dispi_support_bank_granularity_32k:
+  push dx
+  push bx
+  call dispi_get_enable
+  mov  bx, ax
+  or   ax, # VBE_DISPI_GETCAPS
+  call _dispi_set_enable
+  mov  dx, # VBE_DISPI_IOPORT_INDEX
+  mov  ax, # VBE_DISPI_INDEX_BANK
+  out  dx, ax
+  mov  dx, # VBE_DISPI_IOPORT_DATA
+  in   ax, dx
+  push ax
+  mov  ax, bx
+  call _dispi_set_enable
+  pop  ax
+  shr  ax, #8
+  and  al, #VBE_DISPI_BANK_GRANULARITY_32K
+  pop  bx
+  pop  dx
+  ret
+
 _dispi_set_enable:
   push dx
   push ax
@@ -383,7 +419,15 @@ dispi_get_enable:
   pop  dx
   ret
 
-_dispi_set_bank:
+dispi_set_bank:
+  test bl, #0x01
+  jz   no_wr_bank
+  or   ax, #VBE_DISPI_BANK_WR
+no_wr_bank:
+  test bl, #0x02
+  jz   no_rd_bank
+  or   ax, #VBE_DISPI_BANK_RD
+no_rd_bank:
   push dx
   push ax
   mov  dx, # VBE_DISPI_IOPORT_INDEX
@@ -409,35 +453,26 @@ ASM_END
 static void dispi_set_bank_farcall()
 {
 ASM_START
-  cmp bx,#0x0100
+  cmp  bx, #0x0100
   je dispi_set_bank_farcall_get
-  or bx,bx
-  jnz dispi_set_bank_farcall_error
-  mov ax,dx
-  push dx
-  push ax
-  mov ax,# VBE_DISPI_INDEX_BANK
-  mov dx,# VBE_DISPI_IOPORT_INDEX
-  out dx,ax
-  pop ax
-  mov dx,# VBE_DISPI_IOPORT_DATA
-  out dx,ax
-  in  ax,dx
-  pop dx
-  cmp dx,ax
-  jne dispi_set_bank_farcall_error
-  mov ax, #0x004f
+  or   bx, bx
+  jnz  dispi_set_bank_farcall_error
+  mov  ax, dx
+  or   ax, #VBE_DISPI_BANK_RW
+  call dispi_set_bank
+  in   ax, dx
+  pop  dx
+  cmp  dx, ax
+  jne  dispi_set_bank_farcall_error
+  mov  ax, #0x004f
   retf
 dispi_set_bank_farcall_get:
-  mov ax,# VBE_DISPI_INDEX_BANK
-  mov dx,# VBE_DISPI_IOPORT_INDEX
-  out dx,ax
-  mov dx,# VBE_DISPI_IOPORT_DATA
-  in ax,dx
-  mov dx,ax
+  call dispi_get_bank
+  mov  dx, ax
+  mov  ax, #0x004f
   retf
 dispi_set_bank_farcall_error:
-  mov ax,#0x014F
+  mov  ax, #0x014F
   retf
 ASM_END
 }
@@ -667,6 +702,277 @@ bit9_clear:
 vga_compat_end:
   pop  dx
   pop  ax
+
+;; code for 'write character' support in 8bpp graphics modes
+
+;; called from C code
+_vbe_is_8bpp_mode:
+vbe_is_8bpp_mode:
+  call _vbe_has_vbe_display
+  test ax, #0x01
+  jz   no_vbe_8bpp_mode
+  call dispi_get_enable
+  and  ax, # VBE_DISPI_ENABLED
+  jz   no_vbe_8bpp_mode
+  call dispi_get_bpp
+  cmp  al, #0x08
+  je   is_8bpp_mode
+no_vbe_8bpp_mode:
+  xor  ax, ax
+is_8bpp_mode:
+  ret
+
+vbe_8bpp_write_char:
+  push ax
+  push bx
+  push cx
+  push dx
+  push ds
+  push es
+  push si
+  push di
+  xor  ax, ax
+  mov  ds, ax
+  mov  bx, #0x10c ;; INT 0x43
+  mov  si, [bx]
+  mov  ax, [bx+2]
+  mov  ds, ax
+  mov  bx, 14[bp] ;; cheight
+  mov  al, 4[bp]  ;; character
+  mul  bl
+  add  si, ax
+  mov  ax, 8[bp] ;; xcurs
+  shl  ax, #3
+  push ax
+  mov  dx, 10[bp] ;; ycurs
+  or   dx, dx
+  jz   vbe_wc_set_start
+  mov  al, 12[bp] ;; nbcols
+  mov  bl, 14[bp] ;; cheight
+  mul  bl
+  shl  ax, #3
+  mov  bx, dx
+  mul  bx
+vbe_wc_set_start:
+  pop  bx
+  add  ax, bx
+  jnc  vbe_wc_noinc1
+  inc  dx
+vbe_wc_noinc1:
+  shl  dx, #1
+  test ax, #0x8000
+  jz   vbe_wc_noinc2
+  and  ax, #0x7fff
+  or   dx, #0x0001
+vbe_wc_noinc2:
+  mov  di, ax
+  mov  ax, dx
+  mov  bl, #0x01
+  call dispi_set_bank
+  mov  ax, #0xa000
+  mov  es, ax
+  mov  bl, 14[bp] ;; cheight
+  mov  dx, 12[bp] ;; nbcols
+  shl  dx, #3
+vbe_wc_loop1:
+  push di
+  lodsb
+  mov  cx, #0x08
+  cld
+vbe_wc_loop2:
+  shl  al, #1
+  push ax
+  jnc  vbe_wc_set_bkgnd
+  mov  al, 6[bp] ;; attr
+  db    0xa9 ;; skip next opcode (TEST AX, #0xC030)
+vbe_wc_set_bkgnd:
+  xor  al, al
+  stosb
+  pop  ax
+  loop vbe_wc_loop2
+  pop  di
+  dec  bl
+  jz   vbe_wc_end
+  add  di, dx
+  jmp  vbe_wc_loop1
+vbe_wc_end:
+  pop  di
+  pop  si
+  pop  es
+  pop  ds
+  pop  dx
+  pop  cx
+  pop  bx
+  pop  ax
+  ret
+
+vbe_8bpp_copy:
+  push ax
+  push bx
+  push cx
+  push dx
+  push ds
+  push es
+  push si
+  push di
+  mov  ax, 4[bp] ;; xstart
+  shl  ax, #3
+  push ax
+  mov  al, 12[bp] ;; nbcols
+  mov  bl, 14[bp] ;; cheight
+  mul  bl
+  shl  ax, #3
+  mov  cx, ax
+  xor  ax, ax
+  mov  dx, 6[bp] ;; ysrc
+  or   dx, dx
+  jz   vbe_copy_set_start_src
+  mov  ax, cx
+  mov  bx, dx
+  mul  bx
+vbe_copy_set_start_src:
+  pop  bx
+  push bx
+  add  ax, bx
+  jnc  vbe_copy_noinc1s
+  inc  dx
+vbe_copy_noinc1s:
+  shl  dx, #1
+  test ax, #0x8000
+  jz   vbe_copy_noinc2s
+  and  ax, #0x7fff
+  or   dx, #0x0001
+vbe_copy_noinc2s:
+  mov  si, ax
+  mov  ax, dx
+  mov  bl, #0x02
+  call dispi_set_bank
+  xor  ax, ax
+  mov  dx, 8[bp] ;; ydest
+  or   dx, dx
+  jz   vbe_copy_set_start_dst
+  mov  ax, cx
+  mov  bx, dx
+  mul  bx
+vbe_copy_set_start_dst:
+  pop  bx
+  add  ax, bx
+  jnc  vbe_copy_noinc1d
+  inc  dx
+vbe_copy_noinc1d:
+  shl  dx, #1
+  test ax, #0x8000
+  jz   vbe_copy_noinc2d
+  and  ax, #0x7fff
+  or   dx, #0x0001
+vbe_copy_noinc2d:
+  mov  di, ax
+  mov  ax, dx
+  mov  bl, #0x01
+  call dispi_set_bank
+  mov  ax, #0xa000
+  mov  ds, ax
+  mov  es, ax
+  mov  bl, 14[bp] ;; cheight
+  mov  cx, 10[bp] ;; cols
+  shl  cx, #2
+  mov  dx, 12[bp] ;; nbcols
+  shl  dx, #3
+vbe_copy_loop:
+  push cx
+  push si
+  push di
+  cld
+  rep
+    movsw
+  pop  di
+  pop  si
+  pop  cx
+  dec  bl
+  jz   vbe_copy_end
+  add  si, dx
+  add  di, dx
+  jmp  vbe_copy_loop
+vbe_copy_end:
+  pop  di
+  pop  si
+  pop  es
+  pop  ds
+  pop  dx
+  pop  cx
+  pop  bx
+  pop  ax
+  ret
+
+vbe_8bpp_fill:
+  push ax
+  push bx
+  push cx
+  push dx
+  push es
+  push di
+vbe_fill_next_row:
+  mov  ax, 4[bp] ;; xstart
+  shl  ax, #3
+  push ax
+  xor  ax, ax
+  mov  dx, 6[bp] ;; ystart
+  or   dx, dx
+  jz   vbe_fill_set_start
+  mov  al, 12[bp] ;; nbcols
+  mov  bl, 14[bp] ;; cheight
+  mul  bl
+  shl  ax, #3
+  mov  bx, dx
+  mul  bx
+vbe_fill_set_start:
+  pop  bx
+  add  ax, bx
+  jnc  vbe_fill_noinc1
+  inc  dx
+vbe_fill_noinc1:
+  shl  dx, #1
+  test ax, #0x8000
+  jz   vbe_fill_noinc2
+  and  ax, #0x7fff
+  or   dx, #0x0001
+vbe_fill_noinc2:
+  mov  di, ax
+  mov  ax, dx
+  mov  bl, #0x01
+  call dispi_set_bank
+  mov  ax, #0xa000
+  mov  es, ax
+  mov  al, 16[bp] ;; attr
+  mov  ah, al
+  mov  bl, 14[bp] ;; cheight
+  mov  cx, 8[bp] ;; cols
+  shl  cx, #2
+  mov  dx, 12[bp] ;; nbcols
+  shl  dx, #3
+vbe_fill_loop:
+  push cx
+  push di
+  cld
+  rep
+    stosw
+  pop  di
+  pop  cx
+  dec  bl
+  jz   vbe_fill_check_row
+  add  di, dx
+  jmp  vbe_fill_loop
+vbe_fill_check_row:
+  inc  6[bp] ;; ystart
+  dec  10[bp] ;; rows
+  jnz  vbe_fill_next_row
+  pop  di
+  pop  es
+  pop  dx
+  pop  cx
+  pop  bx
+  pop  ax
+  ret
 ASM_END
 
 
@@ -753,7 +1059,7 @@ no_vbe_interface:
 
 vbe_display_info:
   call _vbe_has_vbe_display
-  test ax, ax
+  or   ax, ax
   jz   no_vbe_flag
   mov  ax, #0xc000
   mov  ds, ax
@@ -778,8 +1084,9 @@ lmulul:
   ret
 ASM_END
 
+
 /** Function 00h - Return VBE Controller Information
- * 
+ *
  * Input:
  *              AX      = 4F00h
  *              ES:DI   = Pointer to buffer in which to place VbeInfoBlock structure
@@ -787,7 +1094,7 @@ ASM_END
  *                        the info block is 512 bytes in size)
  * Output:
  *              AX      = VBE Return Status
- * 
+ *
  */
 void vbe_biosfn_return_controller_information(AX, ES, DI)
 Bit16u *AX;Bit16u ES;Bit16u DI;
@@ -911,14 +1218,14 @@ Bit16u *AX;Bit16u ES;Bit16u DI;
 
 
 /** Function 01h - Return VBE Mode Information
- * 
+ *
  * Input:
  *              AX      = 4F01h
  *              CX      = Mode Number
  *              ES:DI   = Pointer to buffer in which to place ModeInfoBlock structure
  * Output:
  *              AX      = VBE Return Status
- * 
+ *
  */
 void vbe_biosfn_return_mode_information(AX, CX, ES, DI)
 Bit16u *AX;Bit16u CX; Bit16u ES;Bit16u DI;
@@ -929,6 +1236,7 @@ Bit16u *AX;Bit16u CX; Bit16u ES;Bit16u DI;
         ModeInfoListItem  *cur_info;
         Boolean           using_lfb;
         Bit16u            lfb_addr;
+        Bit16u            support_bank_gr_32k;
 
 #ifdef DEBUG
         printf("VBE vbe_biosfn_return_mode_information ES%x DI%x CX%x\n",ES,DI,CX);
@@ -958,6 +1266,10 @@ Bit16u *AX;Bit16u CX; Bit16u ES;Bit16u DI;
                 if (lfb_addr > 0) {
                   info.PhysBasePtr = ((Bit32u)lfb_addr << 16);
                 }
+                support_bank_gr_32k = dispi_support_bank_granularity_32k();
+                if (support_bank_gr_32k != 0) {
+                  info.WinGranularity = VBE_DISPI_BANK_GRANULARITY_KB;
+                }
                 if (info.WinAAttributes & VBE_WINDOW_ATTRIBUTE_RELOCATABLE) {
                   info.WinFuncPtr = 0xC0000000UL;
                   *(Bit16u *)&(info.WinFuncPtr) = (Bit16u)(dispi_set_bank_farcall);
@@ -983,108 +1295,115 @@ Bit16u *AX;Bit16u CX; Bit16u ES;Bit16u DI;
 }
 
 /** Function 02h - Set VBE Mode
- * 
+ *
  * Input:
  *              AX      = 4F02h
  *              BX      = Desired Mode to set
- *              ES:DI   = Pointer to CRTCInfoBlock structure
  * Output:
  *              AX      = VBE Return Status
- * 
+ *
  */
-void vbe_biosfn_set_mode(AX, BX, ES, DI)
-Bit16u *AX;Bit16u BX; Bit16u ES;Bit16u DI;
+void vbe_biosfn_set_mode(AX, BX)
+Bit16u *AX;Bit16u BX;
 {
-        Bit16u            ss = get_SS();
-        Bit16u            result;
-        ModeInfoListItem  *cur_info;
-        Boolean           using_lfb;
-        Bit8u             no_clear;
-        Bit8u             lfb_flag;
+    Bit16u            ss = get_SS();
+    Bit16u            result;
+    ModeInfoListItem  *cur_info;
+    Boolean           using_lfb;
+    Bit8u             no_clear;
+    Bit8u             lfb_flag;
+    Bit16u            support_bank_gr_32k;
 
-        using_lfb=((BX & VBE_MODE_LINEAR_FRAME_BUFFER) == VBE_MODE_LINEAR_FRAME_BUFFER);
-        lfb_flag=using_lfb?VBE_DISPI_LFB_ENABLED:0;
-        no_clear=((BX & VBE_MODE_PRESERVE_DISPLAY_MEMORY) == VBE_MODE_PRESERVE_DISPLAY_MEMORY)?VBE_DISPI_NOCLEARMEM:0;
+    using_lfb=((BX & VBE_MODE_LINEAR_FRAME_BUFFER) == VBE_MODE_LINEAR_FRAME_BUFFER);
+    lfb_flag=using_lfb?VBE_DISPI_LFB_ENABLED:0;
+    no_clear=((BX & VBE_MODE_PRESERVE_DISPLAY_MEMORY) == VBE_MODE_PRESERVE_DISPLAY_MEMORY)?VBE_DISPI_NOCLEARMEM:0;
 
-        BX = (BX & 0x1ff);
+    BX = (BX & 0x1ff);
 
-        //result=read_word(ss,AX);
+    // check for non vesa mode
+    if (BX<VBE_MODE_VESA_DEFINED) {
+        Bit8u   mode;
 
-        // check for non vesa mode
-        if (BX<VBE_MODE_VESA_DEFINED)
-        {
-                Bit8u   mode;
+        dispi_set_enable(VBE_DISPI_DISABLED);
+        // call the vgabios in order to set the video mode
+        // this allows for going back to textmode with a VBE call (some applications expect that to work)
 
-                dispi_set_enable(VBE_DISPI_DISABLED);
-                // call the vgabios in order to set the video mode
-                // this allows for going back to textmode with a VBE call (some applications expect that to work)
-
-                mode=(BX & 0xff);
-                biosfn_set_video_mode(mode);
-                result = 0x4f;
+        mode=(BX & 0xff);
+        if ((mode & 0x80) == 0) {
+          if (no_clear != 0) {
+            mode |= 0x80;
+          }
+          biosfn_set_video_mode(mode);
+          result = 0x4f;
+        } else {
+          result = 0x014f;
         }
+    } else {
 
         cur_info = mode_info_find_mode(BX, using_lfb, &cur_info);
 
-        if (cur_info != 0)
-        {
+        if (cur_info != 0) {
 #ifdef DEBUG
-                printf("VBE found mode %x, setting:\n", BX);
-                printf("\txres%x yres%x bpp%x\n",
-                        cur_info->info.XResolution,
-                        cur_info->info.YResolution,
-                        cur_info->info.BitsPerPixel);
+            printf("VBE found mode %x, setting:\n", BX);
+            printf("\txres%x yres%x bpp%x\n",
+                   cur_info->info.XResolution,
+                   cur_info->info.YResolution,
+                   cur_info->info.BitsPerPixel);
 #endif
 
-                // first disable current mode (when switching between vesa modi)
-                dispi_set_enable(VBE_DISPI_DISABLED);
+            // first disable current mode (when switching between vesa modi)
+            dispi_set_enable(VBE_DISPI_DISABLED);
 
-                if (cur_info->info.BitsPerPixel == 4)
-                {
-                  biosfn_set_video_mode(0x6a);
-                }
+            if (cur_info->info.BitsPerPixel == 4) {
+                biosfn_set_video_mode(0x6a);
+            } else if (cur_info->info.BitsPerPixel == 8) {
+                load_dac_palette(3);
+            }
 
-                if (cur_info->info.BitsPerPixel == 8)
-                {
-                  load_dac_palette(3);
-                }
+            dispi_set_bpp(cur_info->info.BitsPerPixel);
+            dispi_set_xres(cur_info->info.XResolution);
+            dispi_set_yres(cur_info->info.YResolution);
+ASM_START
+            mov  ax, #VBE_DISPI_BANK_RW
+            call dispi_set_bank
+ASM_END
+            support_bank_gr_32k = dispi_support_bank_granularity_32k();
+            dispi_set_enable(VBE_DISPI_ENABLED | support_bank_gr_32k |
+                             no_clear | lfb_flag);
+            vga_compat_setup();
 
-                dispi_set_bpp(cur_info->info.BitsPerPixel);
-                dispi_set_xres(cur_info->info.XResolution);
-                dispi_set_yres(cur_info->info.YResolution);
-                dispi_set_bank(0);
-                dispi_set_enable(VBE_DISPI_ENABLED | no_clear | lfb_flag);
-                vga_compat_setup();
+            write_bda_word(BIOSMEM_NB_COLS,cur_info->info.XResolution>>3);
+            write_bda_word(BIOSMEM_NB_ROWS,(cur_info->info.YResolution>>4)-1);
+            write_bda_word(BIOSMEM_CHAR_HEIGHT,16);
+            write_bda_word(BIOSMEM_CURRENT_PAGE,0);
+            write_bda_word(BIOSMEM_CURSOR_POS,0);
+            write_bda_word(BIOSMEM_VBE_MODE,BX);
+            write_bda_byte(BIOSMEM_VIDEO_CTL,(0x60 | no_clear));
+ASM_START
+            SET_INT_VECTOR(0x43, #0xC000, #_vgafont16)
+ASM_END
+            write_bda_byte(BIOSMEM_CURRENT_MODE,0x5f); // FIXME
 
-                write_word(BIOSMEM_SEG,BIOSMEM_VBE_MODE,BX);
-                write_byte(BIOSMEM_SEG,BIOSMEM_VIDEO_CTL,(0x60 | no_clear));
-
-                result = 0x4f;
-        }
-        else
-        {
+            result = 0x4f;
+        } else {
 #ifdef DEBUG
-                printf("VBE *NOT* found mode %x\n" , BX);
+            printf("VBE *NOT* found mode %x\n" , BX);
 #endif
-                result = 0x100;
-
-                // FIXME: redirect non VBE modi to normal VGA bios operation
-                //        (switch back to VGA mode
-                if (BX == 3)
-                        result = 0x4f;
+            result = 0x014f;
         }
+    }
 
-        write_word(ss, AX, result);
+    write_word(ss, AX, result);
 }
 
 /** Function 03h - Return Current VBE Mode
- * 
+ *
  * Input:
  *              AX      = 4F03h
  * Output:
  *              AX      = VBE Return Status
  *              BX      = Current VBE Mode
- * 
+ *
  */
 ASM_START
 vbe_biosfn_return_current_mode:
@@ -1097,6 +1416,7 @@ vbe_biosfn_return_current_mode:
   mov  bx, # BIOSMEM_VBE_MODE
   mov  ax, [bx]
   mov  bx, ax
+  test bx, bx
   jnz  vbe_03_ok
 no_vbe_mode:
   mov  bx, # BIOSMEM_CURRENT_MODE
@@ -1109,67 +1429,8 @@ vbe_03_ok:
   ret
 ASM_END
 
-
-Bit16u vbe_biosfn_read_video_state_size()
-{
-    return 9 * 2;
-}
-
-void vbe_biosfn_save_video_state(ES, BX)
-     Bit16u ES; Bit16u BX;
-{
-    Bit16u enable, i;
-
-    outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_ENABLE);
-    enable = inw(VBE_DISPI_IOPORT_DATA);
-    write_word(ES, BX, enable);
-    BX += 2;
-    if (!(enable & VBE_DISPI_ENABLED)) 
-        return;
-    for(i = VBE_DISPI_INDEX_XRES; i <= VBE_DISPI_INDEX_Y_OFFSET; i++) {
-        if (i != VBE_DISPI_INDEX_ENABLE) {
-            outw(VBE_DISPI_IOPORT_INDEX, i);
-            write_word(ES, BX, inw(VBE_DISPI_IOPORT_DATA));
-            BX += 2;
-        }
-    }
-}
-
-
-void vbe_biosfn_restore_video_state(ES, BX)
-     Bit16u ES; Bit16u BX;
-{
-    Bit16u enable, i;
-
-    enable = read_word(ES, BX);
-    BX += 2;
-
-    if (!(enable & VBE_DISPI_ENABLED)) {
-        outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_ENABLE);
-        outw(VBE_DISPI_IOPORT_DATA, enable);
-    } else {
-        outw(VBE_DISPI_IOPORT_INDEX, VBE_DISPI_INDEX_XRES);
-        outw(VBE_DISPI_IOPORT_DATA, read_word(ES, BX));
-        BX += 2;
-        outw(VBE_DISPI_IOPORT_INDEX, VBE_DISPI_INDEX_YRES);
-        outw(VBE_DISPI_IOPORT_DATA, read_word(ES, BX));
-        BX += 2;
-        outw(VBE_DISPI_IOPORT_INDEX, VBE_DISPI_INDEX_BPP);
-        outw(VBE_DISPI_IOPORT_DATA, read_word(ES, BX));
-        BX += 2;
-        outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_ENABLE);
-        outw(VBE_DISPI_IOPORT_DATA, enable);
-
-        for(i = VBE_DISPI_INDEX_BANK; i <= VBE_DISPI_INDEX_Y_OFFSET; i++) {
-            outw(VBE_DISPI_IOPORT_INDEX, i);
-            outw(VBE_DISPI_IOPORT_DATA, read_word(ES, BX));
-            BX += 2;
-        }
-    }
-}
-
 /** Function 04h - Save/Restore State
- * 
+ *
  * Input:
  *              AX      = 4F04h
  *              DL      = 00h Return Save/Restore State buffer size
@@ -1180,53 +1441,150 @@ void vbe_biosfn_restore_video_state(ES, BX)
  * Output:
  *              AX      = VBE Return Status
  *              BX      = Number of 64-byte blocks to hold the state buffer (if DL=00h)
- * 
+ *
  */
-void vbe_biosfn_save_restore_state(AX, CX, DX, ES, BX)
-Bit16u *AX; Bit16u CX; Bit16u DX; Bit16u ES; Bit16u *BX;
-{
-    Bit16u ss=get_SS();
-    Bit16u result, val;
+ASM_START
+vbe_biosfn_save_restore_state:
+  cmp  dl, #0x01
+  jb   vbe_read_state_size
+  je   vbe_save_state
+  cmp  dl, #0x02
+  je   vbe_restore_state
+  jmp  vbe_unimplemented
+vbe_read_state_size:
+  call read_vga_state_size
+  test cx, #0x08
+  jz   no_vbe_state_size
+  mov  ax, #VBE_DISPI_INDEX_Y_OFFSET
+  dec  ax
+  shl  ax, #1
+  add  bx, ax
+no_vbe_state_size:
+  add  bx, #0x3f
+  shr  bx, #6
+  mov  ax, #0x004f
+  ret
+vbe_save_state:
+  push bx
+  test cx, #0x07
+  jz   no_save_vga_state
+  call save_vga_state
+no_save_vga_state:
+  test cx, #0x08
+  jz   no_save_vbe_state
+  call vbe_biosfn_save_video_state
+no_save_vbe_state:
+  pop  bx
+  mov  ax, #0x004f
+  ret
+vbe_restore_state:
+  push bx
+  test cx, #0x07
+  jz   no_restore_vga_state
+  call restore_vga_state
+no_restore_vga_state:
+  test cx, #0x08
+  jz   no_rest_vbe_state
+  call vbe_biosfn_restore_video_state
+no_rest_vbe_state:
+  pop  bx
+  mov  ax, #0x004f
+  ret
 
-    result = 0x4f;
-    switch(GET_DL()) {
-    case 0x00:
-        val = biosfn_read_video_state_size2(CX);
-#ifdef DEBUG
-        printf("VGA state size=%x\n", val);
-#endif
-        if (CX & 8)
-            val += vbe_biosfn_read_video_state_size();
-        write_word(ss, BX, val);
-        break;
-    case 0x01:
-        val = read_word(ss, BX);
-        val = biosfn_save_video_state(CX, ES, val);
-#ifdef DEBUG
-        printf("VGA save_state offset=%x\n", val);
-#endif
-        if (CX & 8)
-            vbe_biosfn_save_video_state(ES, val);
-        break;
-    case 0x02:
-        val = read_word(ss, BX);
-        val = biosfn_restore_video_state(CX, ES, val);
-#ifdef DEBUG
-        printf("VGA restore_state offset=%x\n", val);
-#endif
-        if (CX & 8)
-            vbe_biosfn_restore_video_state(ES, val);
-        break;
-    default:
-        // function failed
-        result = 0x100;
-        break;
-    }
-    write_word(ss, AX, result);
-}
+vbe_biosfn_save_video_state:
+  call dispi_get_enable
+  seg  es
+  mov  [bx], ax
+  add  bx, #0x02
+  test ax, #VBE_DISPI_ENABLED
+  jnz  vbe_save_full_state
+  ret
+vbe_save_full_state:
+  push cx
+  push dx
+  mov  cx, #VBE_DISPI_INDEX_XRES
+vbe_save_loop:
+  cmp  cx, #VBE_DISPI_INDEX_ENABLE
+  je   vbe_save_next_reg
+  cmp  cx, #VBE_DISPI_INDEX_VIRT_HEIGHT
+  je   vbe_save_next_reg
+  mov  dx, #VBE_DISPI_IOPORT_INDEX
+  mov  ax, cx
+  out  dx, ax
+  mov  dx, #VBE_DISPI_IOPORT_DATA
+  in   ax, dx
+  seg  es
+  mov  [bx], ax
+  add  bx, #0x02
+vbe_save_next_reg:
+  inc  cx
+  cmp  cx, #VBE_DISPI_INDEX_Y_OFFSET
+  jbe  vbe_save_loop
+  pop  dx
+  pop  cx
+  ret
+
+vbe_biosfn_restore_video_state:
+  seg  es
+  mov  ax, [bx]
+  add  bx, #0x02
+  test ax, #VBE_DISPI_ENABLED
+  jnz  vbe_restore_full_state
+  call _dispi_set_enable
+  ret
+vbe_restore_full_state:
+  push ax
+  xor  ax, ax
+  call _dispi_set_enable
+  seg  es
+  mov  ax, [bx]
+  push ax
+  call _dispi_set_xres
+  inc  sp
+  inc  sp
+  add  bx, #0x02
+  seg  es
+  mov  ax, [bx]
+  push ax
+  call _dispi_set_yres
+  inc  sp
+  inc  sp
+  add  bx, #0x02
+  seg  es
+  mov  ax, [bx]
+  push ax
+  call _dispi_set_bpp
+  inc  sp
+  inc  sp
+  add  bx, #0x02
+  pop  ax
+  call _dispi_set_enable
+  call _vga_compat_setup
+  push cx
+  push dx
+  mov  cx, #VBE_DISPI_INDEX_BANK
+vbe_restore_loop:
+  cmp  cx, #VBE_DISPI_INDEX_VIRT_HEIGHT
+  je   vbe_save_next_reg
+  mov  dx, #VBE_DISPI_IOPORT_INDEX
+  mov  ax, cx
+  out  dx, ax
+  mov  dx, #VBE_DISPI_IOPORT_DATA
+  seg  es
+  mov  ax, [bx]
+  out  dx, ax
+  add  bx, #0x02
+vbe_restore_next_reg:
+  inc  cx
+  cmp  cx, #VBE_DISPI_INDEX_Y_OFFSET
+  jbe  vbe_restore_loop
+  pop  dx
+  pop  cx
+  ret
+ASM_END
 
 /** Function 05h - Display Window Control
- * 
+ *
  * Input:
  *              AX      = 4F05h
  *     (16-bit) BH      = 00h Set memory window
@@ -1239,7 +1597,7 @@ Bit16u *AX; Bit16u CX; Bit16u DX; Bit16u ES; Bit16u *BX;
  * Note:
  *              If this function is called while in a linear frame buffer mode,
  *              this function must fail with completion code AH=03h
- * 
+ *
  * Output:
  *              AX      = VBE Return Status
  *              DX      = Window number in window granularity units
@@ -1256,7 +1614,8 @@ vbe_biosfn_display_window_control:
   ret
 set_display_window:
   mov  ax, dx
-  call _dispi_set_bank
+  or   ax, #VBE_DISPI_BANK_RW
+  call dispi_set_bank
   call dispi_get_bank
   cmp  ax, dx
   jne  vbe_05_failed
@@ -1284,13 +1643,13 @@ ASM_END
  *              CX      = If BL=00h Desired Width in Pixels
  *                        If BL=02h Desired Width in Bytes
  *                        (Ignored for Get Functions)
- * 
- * Output: 
+ *
+ * Output:
  *              AX      = VBE Return Status
  *              BX      = Bytes Per Scan Line
  *              CX      = Actual Pixels Per Scan Line
  *                        (truncated to nearest complete pixel)
- *              DX      = Maximum Number of Scan Lines 
+ *              DX      = Maximum Number of Scan Lines
  */
 ASM_START
 vbe_biosfn_set_get_logical_scan_line_length:
@@ -1338,7 +1697,7 @@ ASM_END
 
 
 /** Function 07h - Set/Get Display Start
- * 
+ *
  * Input(16-bit):
  *              AX      = 4F07h
  *              BH      = 00h Reserved and must be 00h
@@ -1365,13 +1724,13 @@ ASM_END
  *                        If BL=04h 0 if flip has not occurred, not 0 if it has
  *              DX      = If BL=01h First Displayed Scan Line
  *
- * Input(32-bit): 
+ * Input(32-bit):
  *              BH      = 00h Reserved and must be 00h
  *              BL      = 00h Set Display Start
  *                      = 80h Set Display Start during Vertical Retrace
  *              CX      = Bits 0-15 of display start address
  *              DX      = Bits 16-31 of display start address
- *              ES      = Selector for memory mapped registers 
+ *              ES      = Selector for memory mapped registers
  */
 ASM_START
 vbe_biosfn_set_get_display_start:
@@ -1398,10 +1757,10 @@ get_display_start:
   mov  ax, #0x004f
   ret
 ASM_END
-  
+
 
 /** Function 08h - Set/Get Dac Palette Format
- * 
+ *
  * Input:
  *              AX      = 4F08h
  *              BL      = 00h set DAC palette width
@@ -1450,6 +1809,7 @@ ASM_END
  *           AX    = 4F09h
  *           BL    = 00h     Set palette data
  *           BL    = 01h     Get palette data
+ *           BL    = 80h     Set palette data during Vertical Retrace
  *           CX    =         Number of palette registers to read/write
  *           DX    =         First of palette registers to read/write
  *           ES:DI =         Pointer to buffer with table of palette values
@@ -1460,10 +1820,12 @@ ASM_END
  */
 ASM_START
 vbe_biosfn_set_get_palette_data:
-  cmp  bl,#0x01
-  jb   vbe_set_palette_data
-  je   vbe_get_palette_data
-  mov ax, #0x024f ; unimplemented
+  cmp   bl, #0x80
+  je    vbe_set_palette_data
+  cmp   bl, #0x01
+  jb    vbe_set_palette_data
+  je    vbe_get_palette_data
+  mov   ax, #0x024f ; unimplemented
   ret
 vbe_set_palette_data:
   mov   al, dl
@@ -1710,32 +2072,32 @@ vbe_biosfn_display_identification_extensions:
   cmp  bl,#0x01
   jb   vbe_edid_get_capabilities
   je   vbe_read_EDID
-  jmp  vbe_edid_unimplemented
+  jmp  vbe_unimplemented
 
 vbe_edid_get_capabilities:
   test cx,cx
-  jne  vbe_edid_unimplemented
+  jne  vbe_unimplemented
   mov  ax, #0x004f
   mov  bx, #0x0202
   ret
 
 vbe_read_EDID:
   call vbe_ddc_init
-  jnz  vbe_edid_unimplemented
+  jnz  vbe_unimplemented
   call vbe_ddc_start
   call vbe_ddc_delay
   mov  al, #0xa0
   call vbe_ddc_send_byte
-  jc   vbe_edid_unimplemented
+  jc   vbe_unimplemented
   mov  al, #0x00
   call vbe_ddc_send_byte
-  jc   vbe_edid_unimplemented
+  jc   vbe_unimplemented
   call vbe_ddc_stop
   call vbe_ddc_start
   call vbe_ddc_delay
   mov  al, #0xa1
   call vbe_ddc_send_byte
-  jnz  vbe_edid_unimplemented
+  jc   vbe_unimplemented
   push cx
   push di
   mov  cx, #0x0080
@@ -1751,7 +2113,73 @@ vbe_read_edid_loop:
   mov  ax, #0x004f
   ret
 
-vbe_edid_unimplemented:
+vbe_unimplemented:
   mov  ax, #0x014F ;; not implemented
   ret
+
+; Handle INT 10h AH 4Fh
+vbe_main_handler:
+  push ax
+  call _vbe_has_vbe_display
+  shr  al, #1
+  pop  ax
+  jnc  vbe_unsupported
+  cmp  al, #0x15
+  jbe  vbe_call_table
+vbe_unsupported:
+  mov  ax, #0x014F ;; unsupported
+  jmp  int10_end
+vbe_call_table:
+  push bx
+  xor  bx, bx
+  mov  bl, al
+  shl  bx, 1
+ db 0x2e ;; cs:
+  mov  bp, vbe_handlers[bx]
+  pop  bx
+  push #int10_end
+  push bp
+  ret
+
+vbe_normal:
+  push es
+  push ds
+  pusha
+  mov   bx, #0xc000
+  mov   ds, bx
+  call _int10_func
+  popa
+  pop ds
+  pop es
+  ret
+
+vbe_handlers:
+  ;; 00h
+  dw vbe_normal
+  dw vbe_normal
+  dw vbe_normal
+  dw vbe_biosfn_return_current_mode
+  ;; 04h
+  dw vbe_biosfn_save_restore_state
+  dw vbe_biosfn_display_window_control
+  dw vbe_biosfn_set_get_logical_scan_line_length
+  dw vbe_biosfn_set_get_display_start
+  ;; 08h
+  dw vbe_biosfn_set_get_dac_palette_format
+  dw vbe_biosfn_set_get_palette_data
+  dw vbe_biosfn_return_protected_mode_interface
+  dw vbe_unimplemented
+  ;; 0Ch
+  dw vbe_unimplemented
+  dw vbe_unimplemented
+  dw vbe_unimplemented
+  dw vbe_unimplemented
+  ;; 10h
+  dw vbe_unimplemented
+  dw vbe_unimplemented
+  dw vbe_unimplemented
+  dw vbe_unimplemented
+  ;; 14h
+  dw vbe_unimplemented
+  dw vbe_biosfn_display_identification_extensions
 ASM_END
