@@ -1,5 +1,5 @@
 /*
- * $Id: vgabios.h 216 2019-12-30 23:40:08Z vruppert $
+ * $Id: vgabios.h 238 2020-06-23 18:09:48Z vruppert $
  */
 
 #ifndef vgabios_h_included
@@ -45,7 +45,5 @@ typedef unsigned short Boolean;
 #define WITH_ATTR   3
 
 #define SCREEN_SIZE(x,y) (((x*y*2)|0x00ff)+1)
-#define SCREEN_MEM_START(x,y,p) ((((x*y*2)|0x00ff)+1)*p)
-#define SCREEN_IO_START(x,y,p) ((((x*y)|0x00ff)+1)*p)
 
 #endif
